# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "hydride"
__author__ = "Patrick Kunzmann"
__all__ = ["estimate_amino_acid_charges"]

import numpy as np
import biotite.structure as struc


PK_NH3 = {
    "ALA" :  9.71,
    "ARG" :  9.00,
    "ASN" :  8.73,
    "ASP" :  9.66,
    "CYS" : 10.28,
    "GLU" :  9.58,
    "GLN" :  9.00,
    "GLY" :  9.58,
    "HIS" :  9.09,
    "ILE" :  9.60,
    "LEU" :  9.58,
    "LYS" :  9.16,
    "MET" :  9.08,
    "PHE" :  9.09,
    "PRO" : 10.47,
    "SER" :  9.05,
    "THR" :  8.96,
    "TRP" :  9.34,
    "TYR" :  9.04,
    "VAL" :  9.52,
}

PK_COOH = {
    "ALA" : 2.33,
    "ARG" : 2.03,
    "ASN" : 2.16,
    "ASP" : 1.95,
    "CYS" : 1.91,
    "GLU" : 2.16,
    "GLN" : 2.18,
    "GLY" : 2.34,
    "HIS" : 1.70,
    "ILE" : 2.26,
    "LEU" : 2.32,
    "LYS" : 2.15,
    "MET" : 2.16,
    "PHE" : 2.18,
    "PRO" : 1.95,
    "SER" : 2.13,
    "THR" : 2.20,
    "TRP" : 2.38,
    "TYR" : 2.24,
    "VAL" : 2.27,
}

# pK values for side chains
# The second value gives the expected charge if pH is less than pK
PK_SIDE_CHAIN = {
    ("ARG", "NH2") : (12.10, 1),
    ("ASP", "OD2") : ( 3.71, 0),
    ("CYS",  "SG") : ( 8.14, 0),
    ("GLU", "OE2") : ( 4.15, 0),
    ("HIS", "ND1") : ( 6.04, 1),
    ("LYS",  "NZ") : (10.67, 1),
    ("TYR",  "OH") : (10.10, 0),
}


def estimate_amino_acid_charges(atoms, ph):
    """
    Estimate the charge of heavy atoms in peptides based on the
    protonation state of the free amino acid [1]_.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        The atoms to calculate the charges for.
    ph : float
        The charges are estimated based on this *pH* value.
    
    Returns
    -------
    charges : ndarray, shape=(n,), dtype=int
        The estimated charges.
        0 for all atoms that are not part of an amino acid.
    
    References
    ----------
    
    .. [1] DR Lide,
       "CRC Handbook of Chemistry and Physics."
       CRC Press, (2003).
    """
    charges_nh3 = {
        res_name : 1 if ph < pk else 0
        for res_name, pk in PK_NH3.items()
    }
    charges_cooh = {
        res_name : 0 if ph < pk else -1
        for res_name, pk in PK_COOH.items()
    }
    charges_side_chain = {
        key : charge if ph < pk else charge - 1
        for key, (pk, charge) in PK_SIDE_CHAIN.items()
    }

    atom_charges = np.zeros(atoms.array_length(), dtype=int)
    
    # Charges for termini
    amino_acid_mask = struc.filter_amino_acids(atoms)
    amino_indices   = np.where((atoms.atom_name ==   "N") & amino_acid_mask)[0]
    carboxy_indices = np.where((atoms.atom_name == "OXT") & amino_acid_mask)[0]
    chain_starts = struc.get_chain_starts(atoms, add_exclusive_stop=True)
    for i in range(len(chain_starts) -1):
        start = chain_starts[i]
        stop  = chain_starts[i+1]
        chain_amino_indices = amino_indices[
            (amino_indices >= start) & (amino_indices < stop)
        ]
        chain_carboxy_indices = carboxy_indices[
            (carboxy_indices >= start) & (carboxy_indices < stop)
        ]
        if len(chain_amino_indices) > 0:
            amino_i = np.min(chain_amino_indices)
            atom_charges[amino_i] = charges_nh3[atoms.res_name[amino_i]]
        if len(chain_carboxy_indices) > 0:
            carboxy_i = np.max(chain_carboxy_indices)
            atom_charges[carboxy_i] = charges_cooh[atoms.res_name[carboxy_i]]
    
    # Charges for other heavy atoms
    for (res_name, atom_name), charge in charges_side_chain.items():
        mask  = (atoms.res_name == res_name) & (atoms.atom_name == atom_name)
        atom_charges[mask] = charge

    return atom_charges


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import numpy as np
import biotite.structure.info as info
import hydride


@pytest.mark.parametrize(
    "heavy_atom_name, ref_hydrogen_names",
    [
        ("N", ["H", "H1", "H2"]),
        ("CB", ["HB", "HB1", "HB2"]),
        ("C42", ["H42", "H42A", "H42B"]),
    ]
)
def test_generated_names(heavy_atom_name, ref_hydrogen_names):
    """
    Test correct generation of hydrogen names for atoms that are not in
    the library, based on known examples.
    """
    name_lib = hydride.AtomNameLibrary()
    gen = name_lib.generate_hydrogen_names("", heavy_atom_name)
    test_hydrogen_names = [next(gen) for _ in range(len(ref_hydrogen_names))]
    
    assert test_hydrogen_names == ref_hydrogen_names


np.random.seed(0)
res_names = info.all_residues()
@pytest.mark.parametrize(
    "res_name",
    [res_names[i] for i in
     np.random.choice(np.arange(len(res_names)), size=1000, replace=False)]
)
def test_names_from_library(res_name):
    """
    Test correct lookup of hydrogen names for atoms that are in the
    library.
    Use a random selection of molecules from the CCD,
    and check whether the created hydrogen names are equal to the
    existing ones.
    """
    try:
        residue = info.residue(res_name)
    except KeyError:
        pytest.skip("No structure available for residue name")
    name_lib = hydride.AtomNameLibrary()
    name_lib.add_molecule(residue)
    for i in np.where(residue.element != "H")[0]:
        bond_indices, _ = residue.bonds.get_bonds(i)
        ref_hydrogen_names = [
            residue.atom_name[j] for j in bond_indices
            if residue.element[j] == "H"
        ]
        
        gen = name_lib.generate_hydrogen_names(res_name, residue.atom_name[i])
        test_hydrogen_names = [next(gen) for _ in range(len(ref_hydrogen_names))]

        assert test_hydrogen_names == ref_hydrogen_names


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "hydride"
__author__ = "Patrick Kunzmann"

from os.path import splitext
import sys
import argparse
import warnings
import numpy as np
import biotite.structure as struc
import biotite.structure.io.pdb as pdb
import biotite.structure.io.pdbx as pdbx
import biotite.structure.io.mmtf as mmtf
import biotite.structure.io.mol as mol
from .add import add_hydrogen
from. charge import estimate_amino_acid_charges
from .fragments import FragmentLibrary
from .names import AtomNameLibrary
from .relax import relax_hydrogen


class UserInputError(Exception):
    pass


def main(args=None):
    parser = argparse.ArgumentParser(
        description="This program adds hydrogen atoms to molecular "
                    "structures where these are missing.\n"
                    "For more information, please visit "
                    "https://hydride.biotite-python.org/.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "--infile", "-i", metavar="FILE",
        help="The path to the input structure file containing the model "
              "without hydrogen atoms. "
              "If omitted, the file is read from STDOUT."
    )
    parser.add_argument(
        "--outfile", "-o", metavar="FILE",
        help="The path to the output structure file where the model "
             "with added hydrogen atoms should be written to. "
             "If omitted, the file is written to STDOUT. "
             "Any existing hydrogen atoms will be removed the model."
    )
    parser.add_argument(
        "--informat", "-I",
        choices=["pdb", "pdbx", "cif", "mmtf", "sdf", "mol"],
        help="The file format of the input file. "
             "Must be specified if input file is read from STDIN. "
             "If omitted, the file format is guessed from the suffix "
             "of the file."
    )
    parser.add_argument(
        "--outformat", "-O",
        choices=["pdb", "pdbx", "cif", "mmtf", "mol"],
        help="The file format of the output file. "
             "Must be specified if output file is written to STDOUT."
             "If omitted, the file format is guessed from the suffix "
             "of the file."
    )

    parser.add_argument(
        "--verbose", "-v", action="store_true",
        help="Verbose display of errors."
    )

    parser.add_argument(
        "--no-relax", action="store_true",
        help="Omit the relaxation step. "
             "Note bond lengths and angles will still be correct. "
             "However clashes or electrostatically unfavorable "
             "conformations will not be resolved."
    )
    parser.add_argument(
        "--iterations", "-n", type=int, metavar="NUMBER",
        help="The maximum number of relaxation iterations. "
             "The runtime of the relaxation scales approximately "
             "linear with this value, if the relaxation does not "
             "terminate before. "
             "By default, the relaxation runs until a local optimum "
             "has been reached."
    )
    parser.add_argument(
        "--angle-increment", "-a", type=float, metavar="NUMBER", default=10.0,
        help="The angle in degrees that a freely rotatable bond is rotated "
             "in each relaxation step."
             "Lower values increase the accuracy of hydrogen positioning, "
             "but increase the required number of steps until an optimum "
             "is found."
    )
    parser.add_argument(
        "--fragments", "-f", metavar="FILE", action="append",
        help="Additional structure file to containing fragments for the "
             "fragment library. "
             "This can be used supply fragments for molecules with uncommon "
             "groups, if the standard fragment library does not contain such "
             "fragments, yet. "
             "May be supplied multiple times."
    )
    parser.add_argument(
        "--fragformat", "-F",
        choices=["pdb", "pdbx", "cif", "mmtf", "sdf", "mol"],
        help="The file format of the additional structure files. "
             "If omitted, the file format is guessed from the suffix "
             "of the file."
    )
    parser.add_argument(
        "--ignore", "-g", metavar="RESIDUE", action="append", nargs=2,
        help="No hydrogen atoms are added to the specified residue. "
             "The format is '{chain} {residue}', e.g. 'A 123'. "
             "May be supplied multiple times, if multiple residues should be "
             "ignored."
    )
    parser.add_argument(
        "--model", "-m", type=int, metavar="NUMBER", default=1,
        help="The model number, if the input structure file contains multiple "
             "models."
    )
    parser.add_argument(
        "--charges", "-c", type=float, metavar="PH",
        help="Recalculate the charges of atoms in amino acids based on the "
             "given pH value."
             "This estimation does not take the surrounding amino acids into "
             "account."
    )
    
    args = parser.parse_args(args=args)


    try:
        run(args)
    except UserInputError as e:
        print(f"Error: {e}", file=sys.stderr)
        if args.verbose:
            raise
        else:
            sys.exit(1)
    except Exception as e:
        print("An unexpected error occured:\n", file=sys.stderr)
        raise
    


def run(args):
    frag_library = FragmentLibrary.standard_library()
    name_library = AtomNameLibrary.standard_library()
    if args.fragments is not None:
        for frag_path in args.fragments:
            try:
                model = read_structure(frag_path, args.fragformat, 1)
            except UserInputError:
                raise
            except PermissionError:
                raise UserInputError(
                    f"Missing file permission for reading '{frag_path}'"
                )
            except FileNotFoundError:
                raise UserInputError(
                    f"Input file '{args.infile}' cannot be found"
                )
            except:
                raise UserInputError(
                    f"Input file '{args.infile}' contains invalid data"
                )
            frag_library.add_molecule(model)
            name_library.add_molecule(model)
    
    try:
        model = read_structure(args.infile, args.informat, args.model)
    except UserInputError:
        raise
    except PermissionError:
        raise UserInputError(
            f"Missing file permission for reading '{args.infile}'"
        )
    except FileNotFoundError:
        raise UserInputError(
            f"Input file '{args.infile}' cannot be found"
        )
    except:
        raise UserInputError(
            f"Input file '{args.infile}' contains invalid data"
        )
    
    heavy_mask = (model.element != "H")
    if not heavy_mask.all():
        warnings.warn("Existing hydrogen atoms were removed")
        model = model[heavy_mask]
        pass
    
    if args.charges:
        aa_mask = struc.filter_amino_acids(model)
        charges = estimate_amino_acid_charges(model, args.charges)
        model.charge[aa_mask] = charges[aa_mask]
    
    input_mask = np.ones(model.array_length(), dtype=bool)
    if args.ignore is not None:
        for chain_id, res_id in args.ignore:
            res_id = int(res_id)
            removal_mask = (model.chain_id == chain_id) & \
                           (model.res_id   == res_id)
            if not removal_mask.any():
                raise UserInputError(
                    f"Cannot find '{chain_id} {res_id}' "
                    "in the input structure"
                )
            input_mask &= ~removal_mask
    
    model, _ = add_hydrogen(model, input_mask, frag_library, name_library)
    if not args.no_relax:
        if args.iterations is not None and args.iterations < 0:
            raise UserInputError("The number of iterations must be positive")
        model.coord = relax_hydrogen(
            model, args.iterations,
            angle_increment=np.deg2rad(args.angle_increment)
        )
    
    try:
        write_structure(args.outfile, args.outformat, model)
    except UserInputError:
        raise
    except PermissionError:
        raise UserInputError(
            f"Missing file permission for writing '{args.outfile}'"
        )
    except:
        raise


def read_structure(path, format, model_number):
    if format is None:
        if path is None:
            raise UserInputError(
                "The input file format must be given, "
                "if the input file is read from STDIN"
            )
        format = guess_format(path)
    
    if model_number < 1:
        raise UserInputError("Model number must be positive")
    
    if path is None:
        path = sys.stdin
    
    if format == "pdb":
        pdb_file = pdb.PDBFile.read(path)
        model_count = pdb.get_model_count(pdb_file)
        if model_number > model_count:
            raise UserInputError(
                f"Model number {model_number} is out of range "
                f"for the input structure with {model_count} models"
            )
        model = pdb.get_structure(
            pdb_file, model=model_number, extra_fields=["charge"]
        )
        model.bonds = struc.connect_via_residue_names(model)
    elif format == "pdbx" or format == "cif":
        pdbx_file = pdbx.PDBxFile.read(path)
        model_count = pdbx.get_model_count(pdbx_file)
        if model_number > model_count:
            raise UserInputError(
                f"Model number {model_number} is out of range "
                f"for the input structure with {model_count} models"
            )
        model = pdbx.get_structure(
            pdbx_file, model=model_number, extra_fields=["charge"]
        )
        model.bonds = struc.connect_via_residue_names(model)
    elif format == "mmtf":
        if path == sys.stdin:
            # Special handling for binary input
            mmtf_file = mmtf.MMTFFile.read(sys.stdin.buffer)
        else:
            mmtf_file = mmtf.MMTFFile.read(path)
        model_count = mmtf.get_model_count(mmtf_file)
        if model_number > model_count:
            raise UserInputError(
                f"Model number {model_number} is out of range "
                f"for the input structure with {model_count} models"
            )
        model = mmtf.get_structure(
            mmtf_file, model=model_number, include_bonds=True,
            extra_fields=["charge"]
        )
        if model.bonds.get_bond_count() == 0:
            # No bonds were stored in MMTF file
            # -> Predict bonds 
            model.bonds = struc.connect_via_residue_names(model)
    elif format == "mol" or format == "sdf":
        mol_file = mol.MOLFile.read(path)
        if model_number > 1:
            raise UserInputError(
                f"Model number {model_number} is out of range "
                f"for the input structure with 1 models"
            )
        model = mol_file.get_structure()
        model.res_name[:] = mol_file.lines[0].strip()
    else:
        raise UserInputError(f"Unknown file format '{format}'")
    
    return model


def write_structure(path, format, model):
    if format is None:
        if path is None:
            raise UserInputError(
                "The output file format must be given, "
                "if the output written to STDOUT"
            )
        format = guess_format(path)
    
    if path is None:
        path = sys.stdout
    
    if format == "pdb":
        pdb_file = pdb.PDBFile()
        pdb.set_structure(pdb_file, model)
        pdb_file.write(path)
    elif format == "pdbx" or format == "cif":
        pdbx_file = pdbx.PDBxFile()
        pdbx.set_structure(pdbx_file, model, data_block="STRUCTURE")
        pdbx_file.write(path)
    elif format == "mmtf":
        mmtf_file = mmtf.MMTFFile()
        mmtf.set_structure(mmtf_file, model)
        if path == sys.stdout:
            # Special handling for binary output
            mmtf_file.write(sys.stdout.buffer)
        else:
            mmtf_file.write(path)
    elif format == "mol" or format == "sdf":
        mol_file = mol.MOLFile()
        mol_file.set_structure(model)
        mol_file.set_header(model.res_name[0])
        mol_file.write(path)
    else:
        raise UserInputError(f"Unknown file format '{format}'")


def guess_format(path):
    suffix = splitext(path)[-1].lower()
    if suffix in [".pdb"]:
        return "pdb"
    elif suffix in [".pdbx", ".cif", ".mmcif"]:
        return "pdbx"
    elif suffix in [".mmtf"]:
        return "mmtf"
    elif suffix in [".mol", ".sdf"]:
        return "mol"
    else:
        raise UserInputError(f"Unknown file extension '{suffix}'")


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "hydride"
__author__ = "Patrick Kunzmann"
__all__ = ["AtomNameLibrary"]

from os.path import join, dirname, abspath
import pickle
import string
import numpy as np
from biotite.structure.error import BadStructureError


class AtomNameLibrary:
    """
    A library for generating hydrogen atom names.

    For each molecule added to the :class:`AtomNameLibrary`,
    the hydrogen atom names are saved for each heavy atom in this
    molecule.

    If hydrogen atom names should be generated for a heavy atom,
    the library first looks for a corresponding entry in the library.
    If such entry is not found, since the molecule was never added to
    the library, the hydrogen atom names are guessed based on common
    hydrogen naming schemes.
    """

    _std_library = None

    def __init__(self):
        self._name_dict = {}
    

    @staticmethod
    def standard_library():
        """
        Get the standard :class:`AtomNameLibrary`.
        The library contains atom names for the most prominent molecules
        including amino acids and nucleotides.

        Returns
        -------
        library : AtomNameLibrary
            The standard library.
        """
        if AtomNameLibrary._std_library is None:
            AtomNameLibrary._std_library = AtomNameLibrary()
            file_name = join(dirname(abspath(__file__)), "names.pickle")
            with open(file_name, "rb") as names_file:
                AtomNameLibrary._std_library._name_dict \
                    = pickle.load(names_file)
        return AtomNameLibrary._std_library
    

    def add_molecule(self, molecule):
        """
        Add the hydrogen atom names for each heavy atom in the molecule
        to the library.

        Parameters
        ----------
        molecule : AtomArray
            The molecule to use the hydrogen atom names from.
        """
        if molecule.bonds is None:
            raise BadStructureError(
                "The input molecule must have an associated BondList"
            )
        
        all_bond_indices, _ = molecule.bonds.get_all_bonds()

        for i in np.where(molecule.element != "H")[0]:
            bonded_indices = all_bond_indices[i]
            # Remove padding values
            bonded_indices = bonded_indices[bonded_indices != -1]
            # Set atom names of bonded hydrogen atoms as values
            hydrogen_names = [
                molecule.atom_name[j] for j in bonded_indices
                if molecule.element[j] == "H"
            ]
            if len(hydrogen_names) > 0:
                self._name_dict[
                    (molecule.res_name[i], molecule.atom_name[i])
                ] = hydrogen_names

        
    def generate_hydrogen_names(self, heavy_res_name, heavy_atom_name):
        """
        Generate hydrogen atom names for the given residue and heavy
        atom name.

        If the residue is not found in the library, the hydrogen atom
        name is guessed based on common hydrogen naming schemes.
        """
        hydrogen_names = self._name_dict.get((heavy_res_name, heavy_atom_name))
        if hydrogen_names is not None:
            # Hydrogen names from library
            for i, hydrogen_name in enumerate(hydrogen_names):
                yield hydrogen_name
            try:
                base_name = hydrogen_name[:-1]
                number = int(hydrogen_name[-1])
                while True:
                    # Proceed by increasing the atom number
                    # e.g. CB -> HB1, HB2, HB3, ...
                    number += 1
                    yield f"{base_name}{number}"
            except ValueError:
                # Atom name has no number at the end
                # -> simply append number
                number = 0
                while True:
                    number += 1
                    yield f"{hydrogen_name}{number}"

        
        else:
            if len(heavy_atom_name) == 0:
                # Atom array has no atom names
                # (loaded e.g. from MOL file)
                # -> Also no atom names for hydrogen atoms
                while True:
                    yield ""
            if heavy_atom_name[-1] in string.digits:
                # Atom name ends with number
                # -> assume ligand atom naming
                # C1 -> H1, H1A, H1B
                number = int(
                    ''.join([c for c in heavy_atom_name if c.isdigit()])
                )
                element = heavy_atom_name[0]
                # C1 -> H1, H1A, H1B
                yield f"H{number}"
                i = 0
                while True:
                    yield f"H{number}{string.ascii_uppercase[i]}"
                    i += 1
            elif len(heavy_atom_name) > 1:
                # e.g. CA -> HA, HA2, HA3, ...
                suffix = heavy_atom_name[1:]
                yield f"H{suffix}"
                number = 1
                while True:
                    yield f"H{suffix}{number}"
                    number += 1

            else:
                # N -> H, H2, H3, ...
                yield "H"
                number = 1
                while True:
                    yield f"H{number}"
                    number += 1
        


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
from biotite.structure.atoms import Atom
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.info as info
import hydride


@pytest.mark.parametrize("res_name", [
    "BNZ", # Benzene
    "BZF", # Benzofuran
    "IND", # indole
    "PZO", # Pyrazole
    "BZI", # Benzimidazole
    "LOM", # Thiazole
    "P1R", # Pyrimidine
    "ISQ", # Isoquinoline
    "NPY", # Naphthalene
    "AN3", # Anthracene
    "0PY", # Pyridine
    "4FT", # Phthalazine
    "URA", # Uracil
    "CHX", # Cyclohexane
    "CEJ", # 1,3-Cyclopentanedione
    "CN",  # Hydrogen cyanide
    "11X", # N-pyridin-3-ylmethylaniline
    "ANL", # Aniline
])
def test_hydrogen_positions(res_name):
    """
    Test whether the assigned hydrogen positions approximately match
    the original hydrogen positions for a given molecule.
    
    All chosen molecules consist completely of heavy atoms without
    rotational freedom for the bonded hydrogen atoms, such as aromatic
    or cyclic compounds.
    This is required to reproduce unambiguously the original hydrogen
    positions, as no relaxation is performed.
    The chosen molecules are also part of the standard fragment library
    themselves.
    However, due to the mere size of the CCD it is improbable that
    a fragment in the library actually originates from the molecule it
    is added to.
    """
    TOLERANCE = 0.1

    library = hydride.FragmentLibrary.standard_library()
    molecule = info.residue(res_name)
    # Perform translation of the molecule along the three axes
    ref_hydrogen_coord = molecule.coord[molecule.element == "H"]
    
    heavy_atoms = molecule[molecule.element != "H"]
    test_hydrogen_coord = library.calculate_hydrogen_coord(heavy_atoms)
    
    test_count = 0
    for coord in test_hydrogen_coord:
        test_count += len(coord)
    assert test_count == np.count_nonzero(molecule.element == "H")
    assert len(test_hydrogen_coord) == heavy_atoms.array_length()
    ref_index = 0
    for i, hydrogen_coord in enumerate(test_hydrogen_coord):
        try:
            # The coord for all hydrogens bonded to the same heavy atom
            if len(hydrogen_coord) == 0:
                # No bonded hydrogen atoms -> nothing to compare
                continue
            elif len(hydrogen_coord) == 1:
                # Only a single hydrogen atom
                # -> unambiguous assignment to reference hydrogen coord
                assert np.max(struc.distance(
                    hydrogen_coord,
                    ref_hydrogen_coord[ref_index : ref_index+1]
                )) <= TOLERANCE
                ref_index += 1
            elif len(hydrogen_coord) == 2:
                # Heavy atom has 2 hydrogen atoms
                # -> Since the hydrogen atoms are indistinguishable,
                # there are two possible assignment to reference
                # hydrogen atoms
                best_distance = min([
                    np.max(struc.distance(
                        hydrogen_coord,
                        ref_hydrogen_coord[ref_index : ref_index+2][::order]
                    ))
                    for order in (1, -1)
                ])
                assert best_distance <= TOLERANCE
                ref_index += 2
            else:
                # Heavy atom has 3 hydrogen atoms
                # -> there is rotational freedom
                # -> invalid test case
                raise ValueError("Invalid test case")
        except AssertionError:
            print(f"Failing central atom: {heavy_atoms.atom_name[i]}")
            raise



def test_missing_fragment():
    """
    If a molecule contains an unknown fragment, check if a warning is
    raised and all other atoms are still hydrogenated.
    """
    TOLERANCE = 0.1

    lib = hydride.FragmentLibrary.standard_library()

    ref_mol = info.residue("BZI") # Benzimidazole
    # It should not be possible to have a nitrogen at this position
    # with a positive charge
    ref_mol.charge[0] = 1
    ref_hydrogen_coord = ref_mol.coord[ref_mol.element == "H"]


    test_mol = test_mol = ref_mol[ref_mol.element != "H"]
    with pytest.warns(
        UserWarning, match="Missing fragment for atom 'N1' at position 0"
    ):
        hydrogen_coord = lib.calculate_hydrogen_coord(test_mol)
    print(hydrogen_coord)
    flattend_coord = []
    for coord in hydrogen_coord:
        flattend_coord += coord.tolist()
    test_hydrogen_coord = np.array(flattend_coord)
    
    assert np.max(struc.distance(
        test_hydrogen_coord,
        # Expect missing first hydrogen due to missing fragment
        ref_hydrogen_coord[1:]
    )) <= TOLERANCE


@pytest.mark.parametrize(
    "lib_enantiomer, subject_enantiomer",
    itertools.product(
        # L-alanine and D-alanine
        ["ALA", "DAL"],
        ["ALA", "DAL"],
    )
)
def test_stereocenter(lib_enantiomer, subject_enantiomer):
    """
    Test whether one enantiomer in the library is sufficient to
    add hydrogen for both enatiomers
    """
    TOLERANCE = 0.1

    lib = hydride.FragmentLibrary()
    lib.add_molecule(info.residue(lib_enantiomer))

    ref_model = info.residue(subject_enantiomer)

    test_model = ref_model[ref_model.element != "H"]
    # As the test case is constructed with the exact same molecule
    # can be in the library, move the molecule to assure that
    # correct hydrogen position calculation is not an artifact
    np.random.seed(0)
    test_model = struc.rotate(test_model, np.random.rand(3))
    test_model = struc.translate(test_model, np.random.rand(3))
    test_model, _ = hydride.add_hydrogen(test_model, fragment_library=lib)

    test_model, _ = struc.superimpose(
        ref_model, test_model, atom_mask = (test_model.element != "H")
    )
    ref_stereo_h_coord = ref_model.coord[ref_model.atom_name == "HA"][0]
    test_stereo_h_coord = test_model.coord[test_model.atom_name == "HA"][0]

    assert struc.distance(test_stereo_h_coord, ref_stereo_h_coord) <= TOLERANCE


@pytest.mark.parametrize("res_name, nitrogen_index", [
    ("ARG",  7), # Arginine NE
    ("ARG",  9), # Arginine NH1
    ("ARG", 10), # Arginine NH2
    ("PRO",  0), # Proline N
    ("ASN",  7), # Asparagine N
    ("T",   15), # Thymidine N3
])
def test_partial_double_bonds(res_name, nitrogen_index):
    """
    It is difficult to assign hydrogen atoms to nitrogen properly,
    as the the geometry can be trigonal planar instead of tetrahedral,
    due to partial double bonds.

    This test checks whether the correct geomtry is found for selected
    nitrogen atoms in chosen examples.
    The chosen examples have no intersect with
    :func:`test_hydrogen_positions()`, as those molecules were already
    tested.

    As in :func:`test_hydrogen_positions()`, it is important that the
    respective nitrogen atom has no rotational freedom.
    """
    TOLERANCE = 0.1

    library = hydride.FragmentLibrary.standard_library()

    molecule = info.residue(res_name)
    if molecule.element[nitrogen_index] != "N":
        raise ValueError("Invalid test case")
    
    np.random.seed(0)
    molecule = struc.rotate(molecule, np.random.rand(3))
    molecule = struc.translate(molecule, np.random.rand(3))
    
    bond_indices, _ = molecule.bonds.get_bonds(nitrogen_index)
    bond_h_indices = bond_indices[molecule.element[bond_indices] == "H"]
    ref_hydrogen_coord = molecule.coord[bond_h_indices]

    molecule = molecule[molecule.element != "H"]
    hydrogen_coord = library.calculate_hydrogen_coord(molecule)
    test_hydrogen_coord = hydrogen_coord[nitrogen_index]

    # The coord for all hydrogens bonded to the same heavy atom
    if len(test_hydrogen_coord) == 0:
        # No bonded hydrogen atoms -> nothing to compare
        raise ValueError("Invalid test case")
    elif len(test_hydrogen_coord) == 1:
        # Only a single hydrogen atom
        # -> unambiguous assignment to reference hydrogen coord
        assert np.max(struc.distance(
            test_hydrogen_coord,
            ref_hydrogen_coord
        )) <= TOLERANCE
    elif len(test_hydrogen_coord) == 2:
        # Heavy atom has 2 hydrogen atoms
        # -> Since the hydrogen atoms are indistinguishable,
        # there are two possible assignment to reference
        # hydrogen atoms
        best_distance = min([
            np.max(struc.distance(
                test_hydrogen_coord,
                ref_hydrogen_coord[::order]
            ))
            for order in (1, -1)
        ])
        assert best_distance <= TOLERANCE
    else:
        # Heavy atom has 3 hydrogen atoms
        # -> there is rotational freedom
        # -> invalid test case
        raise ValueError("Invalid test case")


def test_undefined_bond_type():
    """
    Check if the :class:`FragmentLibrary` raises a warning, if atoms
    are bonded with :attr:`BondType.ANY`, and if the corresponding
    fragments are ignored.
    """
    molecule = info.residue("BNZ")
    # Retain the bonds, but remove the bond type
    molecule.bonds = struc.BondList(
        molecule.array_length(),
        # Remove bond type from existing bonds
        molecule.bonds.as_array()[:,:2]
    )

    # Test handing of 'BondType.ANY' in 'add_molecule()'
    library = hydride.FragmentLibrary()
    with pytest.warns(UserWarning) as record:
        library.add_molecule(molecule)
    assert any([
        True if "undefined bond type" in warning.message.args[0] else False
        for warning in record
    ])
    # No fragment should be added
    assert len(library._frag_dict) == 0

    # Test handing of 'BondType.ANY' in 'calculate_hydrogen_coord()'
    heavy_atoms = molecule[molecule.element != "H"]
    library = hydride.FragmentLibrary.standard_library()
    with pytest.warns(UserWarning) as record:
        hydrogen_coord = library.calculate_hydrogen_coord(heavy_atoms)
    assert any([
        True if "undefined bond type" in warning.message.args[0] else False
        for warning in record
    ])
    assert len(hydrogen_coord) == heavy_atoms.array_length()
    for coord in hydrogen_coord:
        # For each heavy atom there should be no added hydrogen
        assert len(coord) == 0


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import shutil
import tempfile
import itertools
import os
from os.path import join, splitext
import subprocess
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
import biotite.structure.io.mol as mol
import hydride
from hydride.cli import main as run_cli
from .util import data_dir


PDB_ID = "1l2y"
PH = 7.0


@pytest.fixture(scope="module", params=["pdb", "pdbx", "cif", "mmtf"])
def input_file(request):
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), f"{PDB_ID}.mmtf"))
    model = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    model = model[model.element != "H"]
    temp_file = tempfile.NamedTemporaryFile(
        "w", suffix=f".{request.param}", delete=False
    )
    strucio.save_structure(temp_file.name, model)
    temp_file.close()

    yield temp_file.name

    os.remove(temp_file.name)


@pytest.fixture(params=["pdb", "pdbx", "cif", "mmtf"])
def output_file(request):
    temp_file = tempfile.NamedTemporaryFile(
        "r", suffix=f".{request.param}", delete=False
    )
    temp_file.close()

    yield temp_file.name

    os.remove(temp_file.name)


@pytest.fixture()
def dummy_output_file():
    """
    This file should never be written, as the CLI run should fail
    before.
    """
    temp_file = tempfile.NamedTemporaryFile(
        "r", suffix=f".pdb", delete=False
    )

    yield temp_file.name

    temp_file.close()
    os.remove(temp_file.name)


def assert_hydrogen_addition(test_file):
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), f"{PDB_ID}.mmtf"))
    ref_model = mmtf.get_structure(mmtf_file, model=1, include_bonds=True)
    ref_model.charge = hydride.estimate_amino_acid_charges(ref_model, PH)
    # Both naming conventions 'H' and 'H1' are used in PDB
    # However, AtomNameLibrary uses 'H'; it is changed for consistency
    ref_model.atom_name[ref_model.atom_name == "H1"] = "H"

    test_model = strucio.load_structure(test_file)

    assert test_model.array_length() == ref_model.array_length()
    assert test_model.atom_name.tolist() == ref_model.atom_name.tolist()




def test_simple(input_file, output_file):
    """
    Test CLI run without optional parameters.
    """
    run_cli([
        "-v",
        "-i", input_file,
        "-o", output_file,
        "-c", str(PH)
    ])

    assert_hydrogen_addition(output_file)


def test_molfile():
    """
    Test usage of MOL/SDF files for input and output.
    """
    mol_file = mol.MOLFile.read(join(data_dir(), "TYR.sdf"))
    ref_model = mol_file.get_structure()
    model = ref_model[ref_model.element != "H"]

    input_file = tempfile.NamedTemporaryFile(
        "w", suffix=".mol", delete=False
    )
    strucio.save_structure(input_file.name, model)
    input_file.close()

    output_file = tempfile.NamedTemporaryFile(
        "r", suffix=".mol", delete=False
    )
    output_file.close()

    run_cli([
        "-v",
        "-i", input_file.name,
        "-o", output_file.name,
    ])

    mol_file = mol.MOLFile.read(output_file.name)
    test_model = mol_file.get_structure()

    os.remove(input_file.name)
    os.remove(output_file.name)

    assert test_model.array_length() == ref_model.array_length()
    assert test_model.element.tolist() == ref_model.element.tolist()


def test_std_in_out(input_file, output_file):
    """
    Test CLI run with input from STDIN and output to STDOUT.
    """
    in_format  = splitext(input_file )[-1][1:]
    out_format = splitext(output_file)[-1][1:]
    with open(input_file, "rb") as file:
        input_file_content = file.read()

    # Use subprocess instead of API function call
    # to be able to use STDIN/STDOUT
    completed_process = subprocess.run(
        [
            "hydride",
            "-v",
            "-I", in_format,
            "-O", out_format,
            "-c", str(PH)
        ],
        input=input_file_content,
        capture_output=True
    )
    stdout = completed_process.stderr.decode("UTF-8")
    if "ModuleNotFoundError: No module named 'hydride.relax'" in stdout:
        pytest.skip("Relax module is only compiled via pyximport")
    assert completed_process.returncode == 0

    with open(output_file, "wb") as file:
        file.write(completed_process.stdout)
    
    assert_hydrogen_addition(output_file)


@pytest.mark.parametrize("res_name", ["URA"])
def test_extra_fragments(output_file, res_name):
    """
    If the fragments are given for the molecule, where hydrogens atoms
    should be added, the atom names should fit exactly and the hydrogen
    coordinates should fit almost exactly.
    """
    TOLERANCE = 0.01

    ref_molecule = info.residue(res_name)
    frag_temp_file = tempfile.NamedTemporaryFile(
        "w", suffix=f".mmtf", delete=False
    )
    # As the test case is constructed with the exact same molecule
    # can be in the library, move the molecule to assure that
    # correct hydrogen position calculation is not an artifact
    np.random.seed(0)
    frag_molecule = struc.rotate(ref_molecule, np.random.rand(3))
    frag_molecule = struc.translate(frag_molecule, np.random.rand(3))
    strucio.save_structure(frag_temp_file.name, frag_molecule)

    heavy_atoms = ref_molecule[ref_molecule.element != "H"]
    input_temp_file = tempfile.NamedTemporaryFile(
        "w", suffix=f".mmtf", delete=False
    )
    strucio.save_structure(input_temp_file.name, heavy_atoms)

    run_cli([
        "-v",
        "-i", input_temp_file.name,
        "-f", frag_temp_file.name,
        "-o", output_file,
    ])

    frag_temp_file.close()
    os.remove(frag_temp_file.name)
    input_temp_file.close()
    os.remove(input_temp_file.name)

    test_molecule = strucio.load_structure(output_file)
    
    assert test_molecule.array_length() == ref_molecule.array_length()
    # Atoms are added to AtomNameLibrary
    # -> atom names should fit exactly
    assert test_molecule.atom_name.tolist() == ref_molecule.atom_name.tolist()
    # The fragments are from the same molecule
    # -> The hydrogen positions should be almost fit exactly
    assert np.max(struc.distance(test_molecule, ref_molecule)) <= TOLERANCE


def test_ignored_residues(input_file, output_file):
    """
    Test if ignored residues have no associated hydrogen atoms.
    """
    ignored_residues = (5, 13)
    run_cli(
        [
            "-v",
            "-i", input_file,
            "-g", "A", "5",
            "-g", "A", "13",
            "-o", output_file,
        ]
        + list(itertools.chain(
            *[("-g", "A", str(res_id)) for res_id in ignored_residues]
        ))
    )

    test_model = strucio.load_structure(output_file)

    # There should be added hydrogen atoms in the file...
    assert (test_model.element == "H").any()
    # ... but not for the ignored residues
    for res_id in ignored_residues:
        assert not (
            (test_model.element == "H") & (test_model.res_id == res_id)
        ).any()


def test_limited_iterations(input_file, output_file):
    """
    Test CLI run with ``--iterations`` parameter.
    """
    run_cli([
        "-v",
        "-i", input_file,
        "-o", output_file,
        "-c", str(PH),
        "--iterations", str(100)
    ])

    assert_hydrogen_addition(output_file)


def test_angle_increment(input_file, output_file):
    """
    Test CLI run with ``--angle_increment`` parameter.
    """
    run_cli([
        "-v",
        "-i", input_file,
        "-o", output_file,
        "-c", str(PH),
        "--angle-increment", str(5)
    ])

    assert_hydrogen_addition(output_file)


def test_invalid_iteratons(input_file, dummy_output_file):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", input_file,
            "-o", dummy_output_file,
            "-n", str(-1)
        ])
    assert wrapped_exception.value.code == 1


@pytest.mark.parametrize("model", [-10, -1, 0, 39, 100])
def test_invalid_model(input_file, dummy_output_file, model):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", input_file,
            "-o", dummy_output_file,
            "-m", str(model)
        ])
    assert wrapped_exception.value.code == 1


def test_unknown_input_format(input_file, dummy_output_file):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", input_file,
            "-o", dummy_output_file,
            "-I", "abc"
        ])
    assert wrapped_exception.value.code == 2


def test_unknown_output_format(input_file, dummy_output_file):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", input_file,
            "-o", dummy_output_file,
            "-O", "abc"
        ])
    assert wrapped_exception.value.code == 2


def test_unknown_extension(input_file, dummy_output_file):
    temp_file = tempfile.NamedTemporaryFile(
        "w", suffix=".abc", delete=False
    )
    shutil.copy(input_file, temp_file.name)
    
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", temp_file.name,
            "-o", dummy_output_file,
        ])
    assert wrapped_exception.value.code == 1

    temp_file.close()
    os.remove(temp_file.name)


def test_noexisting_file(dummy_output_file):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", "/not/a/file.pdb",
            "-o", dummy_output_file,
        ])
    assert wrapped_exception.value.code == 1


def test_invalid_file_data(input_file, dummy_output_file):
    in_format = splitext(input_file)[-1]
    temp_file = tempfile.NamedTemporaryFile(
        "wb", suffix=in_format, delete=False
    )
    temp_file.write(b"ATOM Invalid data")

    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", temp_file.name,
            "-o", dummy_output_file,
        ])
    assert wrapped_exception.value.code == 1


def test_missing_ignore_residue(input_file, dummy_output_file):
    with pytest.raises(SystemExit) as wrapped_exception:
        run_cli([
            "-i", input_file,
            "-o", dummy_output_file,
            "-g", "A", "42"
        ])
    assert wrapped_exception.value.code == 1


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
from os.path import join
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
import hydride
from .util import data_dir


@pytest.mark.parametrize("res_name", [
    "BNZ", # Benzene
    "BZF", # Benzofuran
    "IND", # indole
    "PZO", # Pyrazole
    "BZI", # Benzimidazole
    "LOM", # Thiazole
    "P1R", # Pyrimidine
    "ISQ", # Isoquinoline
    "NPY", # Naphthalene
    "AN3", # Anthracene
    "0PY", # Pyridine
    "4FT", # Phthalazine
    "URA", # Uracil
    "CHX", # Cyclohexane
    "CEJ", # 1,3-Cyclopentanedione
    "CN",  # Hydrogen cyanide
    "11X", # N-pyridin-3-ylmethylaniline
    "ANL", # Aniline
])
def test_hydrogen_positions(res_name):
    """
    Test whether the assigned hydrogen positions approximately match
    the original hydrogen positions for a given molecule.
    
    All chosen molecules consist completely of heavy atoms without
    rotational freedom for the bonded hydrogen atoms, such as aromatic
    or cyclic compounds.
    This is required to reproduce unambiguously the original hydrogen
    positions, as no relaxation is performed.
    The chosen molecules are also part of the standard fragment library
    themselves.
    However, due to the mere size of the CCD it is improbable that
    a fragment in the library actually originates from the molecule it
    is added to.
    """
    TOLERANCE = 0.1

    library = hydride.FragmentLibrary.standard_library()
    ref_molecule = info.residue(res_name)

    test_molecule = ref_molecule[ref_molecule.element != "H"]
    test_molecule, _ = hydride.add_hydrogen(test_molecule)

    for category in ref_molecule.get_annotation_categories():
        if category == "atom_name":
            # Atom names are tested separately
            continue
        try:
            assert np.all(
                test_molecule.get_annotation(category) ==
                ref_molecule.get_annotation(category)
            )
        except AssertionError:
            print("Failing category:", category)
            raise
    # Atom names are only guessed
    # -> simply check if the atoms names are unique
    assert len(np.unique(test_molecule.atom_name)) \
        == len(test_molecule.atom_name)
    
    for heavy_i in np.where(ref_molecule.element != "H")[0]:
        ref_bond_i, _ = ref_molecule.bonds.get_bonds(heavy_i)
        ref_h_indices = ref_bond_i[ref_molecule.element[ref_bond_i] == "H"]
        test_bond_i, _ = test_molecule.bonds.get_bonds(heavy_i)
        test_h_indices = test_bond_i[test_molecule.element[test_bond_i] == "H"]
        # The coord for all hydrogens bonded to the same heavy atom
        if len(ref_h_indices) == 0:
            # No bonded hydrogen atoms -> nothing to compare
            continue
        elif len(ref_h_indices) == 1:
            # Only a single hydrogen atom
            # -> unambiguous assignment to reference hydrogen coord
            assert np.max(struc.distance(
                test_molecule.coord[test_h_indices],
                ref_molecule.coord[ref_h_indices],
            )) <= TOLERANCE
        elif len(ref_h_indices) == 2:
            # Heavy atom has 2 hydrogen atoms
            # -> Since the hydrogen atoms are indistinguishable,
            # there are two possible assignment to reference hydrogens
            best_distance = min([
                np.max(struc.distance(
                    test_molecule.coord[test_h_indices],
                    ref_molecule.coord[ref_h_indices][::order],
                ))
                for order in (1, -1)
            ])
            assert best_distance <= TOLERANCE
        else:
            # Heavy atom has 3 hydrogen atoms
            # -> there is rotational freedom
            # -> invalid test case
            raise ValueError("Invalid test case")


def test_molecule_without_hydrogens():
    """
    Test whether the :func:`add_hydrogen()` can handle molecules, where
    no hydrogen atom should be added.
    It is expected that simply the original structure is returned
    """
    # Chlordecone
    ref_molecule = info.residue("27E")

    test_molecule, original_mask = hydride.add_hydrogen(ref_molecule)

    assert np.all(original_mask)
    assert test_molecule == ref_molecule


def test_atom_mask():
    """
    Test atom mask usage by hydrogenating in two steps:
    Firstly one random half is masked, secondly the other half is
    masked.
    After reordering the atoms in the standard way, the result should be
    equal to hydrogenating in a single step.
    """
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    heavy_atoms = atoms[atoms.element != "H"]

    ref_atoms, _ = hydride.add_hydrogen(heavy_atoms)
    ref_atoms = ref_atoms[info.standardize_order(ref_atoms)]

    # Hydrogenate the first random half of the molecule
    random_mask = np.random.choice([False, True], heavy_atoms.array_length())
    half_hydrogenated, orig_mask = hydride.add_hydrogen(
        heavy_atoms, random_mask
    )
    # Hydrogenate the second half by inverting mask
    # Special handling due to additional hydrogen atoms
    # after first hydrogenation
    inv_random_mask = np.zeros(half_hydrogenated.array_length(), bool)
    inv_random_mask[orig_mask] = ~random_mask
    test_atoms, _ = hydride.add_hydrogen(
        half_hydrogenated, inv_random_mask
    )
    test_atoms = test_atoms[info.standardize_order(test_atoms)]

    assert test_atoms == ref_atoms


@pytest.mark.parametrize("fill_value", [False, True])
def test_atom_mask_extreme_case(fill_value):
    """
    Check whether the input atom mask works properly by testing the
    cases, where no atom an each atom is masked, respectively.
    """
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    heavy_atoms = atoms[atoms.element != "H"]
    mask = np.full(heavy_atoms.array_length(), fill_value, dtype=bool)

    if fill_value is True:
        # All hydrogen atom are added
        ref_atoms,  ref_orig_mask  = hydride.add_hydrogen(heavy_atoms)
        test_atoms, test_orig_mask = hydride.add_hydrogen(heavy_atoms, mask)
        assert test_atoms == ref_atoms
        assert np.array_equal(test_orig_mask, ref_orig_mask)
    else:
        # No hydrogen atoms are added
        test_atoms, test_orig_mask = hydride.add_hydrogen(heavy_atoms, mask)
        assert test_atoms == heavy_atoms
        assert (test_orig_mask == True).all()


@pytest.mark.parametrize(
    "path",
    glob.glob(join(data_dir(), "*.mmtf"))
)
def test_original_mask(path):
    """
    Check whether the returned original atom mask is correct
    by applying it to the hydrogenated structure and expect that the
    original structure is restored
    """
    mmtf_file = mmtf.MMTFFile.read(path)
    ref_model = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    ref_model = ref_model[ref_model.element != "H"]
    
    hydrogenated_model, original_mask = hydride.add_hydrogen(ref_model)
    test_model = hydrogenated_model[original_mask]

    assert hydrogenated_model.array_length() > ref_model.array_length()
    assert test_model == ref_model


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__version__ = "1.0.0"
__name__ = "hydride"
__author__ = "Jacob Anter, Patrick Kunzmann"

from .add import *
from .charge import *
from .fragments import *
from .names import *

# Module can only be imported if the C-extension has already been built
try:
    from .relax import *
except ImportError:
    pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join, dirname, realpath


def data_dir():
    return join(dirname(realpath(__file__)), "data")


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import itertools
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
import hydride
from hydride.relax import _find_rotatable_bonds
from .util import data_dir


@pytest.fixture
def ethane():
    # Construct ethane in staggered conformation
    ethane = struc.AtomArray(8)
    ethane.element = np.array(["C", "C", "H", "H", "H", "H", "H", "H"])
    ethane.coord = np.array([
        [-0.756,  0.000,  0.000],
        [ 0.756,  0.000,  0.000],
        [-1.140,  0.659,  0.7845],
        [-1.140,  0.350, -0.9626],
        [-1.140, -1.009,  0.1781],
        [ 1.140, -0.350,  0.9626],
        [ 1.140,  1.009, -0.1781],
        [ 1.140, -0.659, -0.7845],
    ])
    ethane.bonds = struc.BondList(
        8,
        np.array([
            [0, 1, 1],
            [0, 2, 1],
            [0, 3, 1],
            [0, 4, 1],
            [1, 5, 1],
            [1, 6, 1],
            [1, 7, 1],
        ])
    )
    ethane.set_annotation("charge", np.zeros(ethane.array_length(), dtype=int))

    # Check if created ethane is in optimal staggered conformation
    # -> Dihedral angle of 60 degrees
    dihed = struc.dihedral(ethane[2], ethane[0], ethane[1], ethane[5])
    assert np.rad2deg(dihed) % 120 == pytest.approx(60, abs=1)

    return ethane


@pytest.mark.parametrize("seed", np.arange(10))
def test_staggered(ethane, seed):
    """
    :func:`relax_hydrogen()` should be able to restore a staggered
    conformation of ethane from any other conformation.
    """
    # Move the ethane molecule away
    # from the optimal staggered conformation
    np.random.seed(seed)
    angle = np.random.rand() * 2 * np.pi
    ethane.coord[5:] = struc.rotate_about_axis(
        ethane.coord[5:],
        angle = angle,
        axis = ethane.coord[1] - ethane.coord[0],
        support = ethane.coord[0]
    )

    # Check if new conformation ethane is not staggered anymore
    dihed = struc.dihedral(ethane[2], ethane[0], ethane[1], ethane[5])
    assert np.rad2deg(dihed) % 120 != pytest.approx(60, abs=1)

    # Try to restore staggered conformation via relax_hydrogen()
    ethane.coord = hydride.relax_hydrogen(
        ethane,
        # The angle increment must be smaller
        # than the expected accuracy (abs=1)
        angle_increment = np.deg2rad(0.5),
    )

    # Check if staggered conformation is restored
    dihed = struc.dihedral(ethane[2], ethane[0], ethane[1], ethane[5])
    assert np.rad2deg(dihed) % 120 == pytest.approx(60, abs=1)


def test_hydrogen_bonds():
    """
    Check whether the relaxation algorithm is able to restore most of
    the original hydrogen bonds.
    The number of bonds found without relaxation is handled as baseline.
    The residues at the biotin binding pocket of streptavidin (including
    biotin itself) are used as test case.
    """
    PERCENTAGE = 1.0
    RES_IDS = (27, 43, 45, 47, 90, 300)

    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "2rtg.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    atoms = atoms[atoms.chain_id == "B"]
    mask = np.isin(atoms.res_id, RES_IDS)
    ref_num = len(struc.hbond(atoms, mask, mask))
    
    atoms = atoms[atoms.element != "H"]
    atoms, _ = hydride.add_hydrogen(atoms)
    mask = np.isin(atoms.res_id, RES_IDS)
    base_num = len(struc.hbond(atoms, mask, mask))

    atoms.coord = hydride.relax_hydrogen(atoms)
    mask = np.isin(atoms.res_id, RES_IDS)
    test_num = len(struc.hbond(atoms, mask, mask))

    if base_num == ref_num:
        ValueError(
            "Invalid test case, "
            "no further hydrogen bonds can be found via relaxation"
        )
    assert (test_num - base_num) / (ref_num - base_num) >= PERCENTAGE


@pytest.mark.parametrize(
    "res_name, ref_bonds",
    [   
        # Fructopyranose
        ("FRU", [
            ( "O1",  "C1",  True,  ("HO1",)),
            ( "O2",  "C2",  True,  ("HO2",)),
            ( "O3",  "C3",  True,  ("HO3",)),
            ( "O4",  "C4",  True,  ("HO4",)),
            ( "O6",  "C6",  True,  ("HO6",)),
        ]),
        # Arginine with positive side chain
        ("ARG", [
            (  "N",  "CA",  True,  ("H", "H2")),
            ("OXT",   "C",  True,  ("HXT",)),
        ]),
        # Isoleucine
        ("ILE", [
            (  "N",  "CA",  True,  ("H", "H2")),
            ("OXT",   "C",  True,  ("HXT",)),
            ("CG2",  "CB",  True,  ("HG21", "HG22", "HG23")),
            ("CD1", "CG1",  True,  ("HD11", "HD12", "HD13")),
        ]),
        # 1-phenylguanidine
        ("PL0", [
            ( "N3",  "C7",  False, ("HN3",)),
        ]),
        # Water
        ("HOH", [
        ]),
    ]
)
def test_bond_identification(res_name, ref_bonds):
    """
    Test whether rotatable bonds for the relaxation are correctly
    identified based on known molecules.
    """
    molecule = info.residue(res_name)
    rotatable_bonds = _find_rotatable_bonds(molecule)

    ref_bonds = set(ref_bonds)

    assert len(rotatable_bonds) == len(ref_bonds)
    for center_atom_i, bonded_atom_i, is_free, h_indices in rotatable_bonds:
        bond_tuple = (
            molecule.atom_name[center_atom_i],
            molecule.atom_name[bonded_atom_i],
            is_free,
            tuple(np.sort(molecule.atom_name[h_indices]))
        )
        assert bond_tuple in ref_bonds


def test_return_trajectory():
    """
    Test whether the `return_trajectory` parameter works properly.
    It is expected that :func:`relax_hydrogen()` returns multiple
    models.
    """
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )

    traj_coord = hydride.relax_hydrogen(atoms, return_trajectory=True)

    assert traj_coord.ndim == 3
    # Last model in trajectory should be the same result
    # as running 'relax_hydrogen()' without 'return_trajectory=True'
    assert np.array_equal(traj_coord[-1], hydride.relax_hydrogen(atoms))


def test_return_energies():
    """
    Test whether the `return_energies` parameter works properly.
    It is expected that :func:`relax_hydrogen()` returns an array of
    energies.
    """
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )

    _, energies = hydride.relax_hydrogen(
        atoms, return_energies=True
    )
    assert isinstance(energies, np.ndarray)
    # Energies should monotonically decrease
    assert (np.diff(energies) <= 0).all()

    traj_coord, energies = hydride.relax_hydrogen(
        atoms, return_energies=True, return_trajectory=True
    )
    assert len(traj_coord) == len(energies)

    assert traj_coord.ndim == 3
    # Last model in trajectory should be the same result
    # as running 'relax_hydrogen()' without 'return_trajectory=True'
    assert np.array_equal(traj_coord[-1], hydride.relax_hydrogen(atoms))


@pytest.mark.parametrize("repulsive", [False, True])
def test_partial_charges(ethane, repulsive):
    """
    Test whether the `partial_charges` parameter is properly used, by
    giving one hydrogen atom on each carbon atom of ethane an
    unphysical charge, either attractive or repulsive to each other.
    This should give result conformations that strongly deviate from
    the staggered conformation, since the electrostatic term should
    minimize or maximize the distance between these hydrogen atoms,
    respectively.
    """
    if repulsive:
        charges = np.array([0, 0, 1, 0, 0, 1, 0, 0])
    else:
        charges = np.array([0, 0, -1, 0, 0, 1, 0, 0])
    
    ethane.coord = hydride.relax_hydrogen(
        ethane,
        # The angle increment must be smaller
        # than the expected accuracy (abs=1)
        angle_increment = np.deg2rad(0.5),
        partial_charges = charges
    )

    # Check if staggered conformation is restored
    dihed = struc.dihedral(ethane[2], ethane[0], ethane[1], ethane[5])
    if repulsive:
        exp_angle = 180
    else:
        exp_angle = 0
    assert np.rad2deg(dihed) % 360 == pytest.approx(exp_angle, abs=1)


def test_limited_iterations():
    """
    Test whether the `iterations` parameter works properly.
    It is expected that the number of returned models,
    if `return_trajectory, is set to true, is equal to the given number
    of maximum iterations.
    That is only true, if the number of iterations is low enough,
    so that the relaxation does not terminate before.
    """
    ITERATIONS = 4
    
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )

    traj_coord = hydride.relax_hydrogen(
        atoms, ITERATIONS, return_trajectory=True
    )

    assert traj_coord.shape[0] == ITERATIONS


@pytest.mark.parametrize(
    "iterations, return_trajectory, return_energies",
    itertools.product(
        [None, 100],
        [False, True],
        [False, True]
    )
)
def test_shortcut_return(iterations, return_trajectory, return_energies):
    """
    Test whether the shortcut return, that happens if no rotatable bonds
    are found, has the same return types as the regular return.
    Therefore the output types of two molecules, one with and one 
    without rotatable bonds, are compared.
    """
    # Rotatable
    ref_atoms  = info.residue("GLY")
    # Non-rotatable
    test_atoms = info.residue("HOH")

    ref_output = hydride.relax_hydrogen(
        ref_atoms, iterations,
        return_trajectory=return_trajectory, return_energies=return_energies
    )
    test_output = hydride.relax_hydrogen(
        test_atoms, iterations,
        return_trajectory=return_trajectory, return_energies=return_energies
    )

    if isinstance(ref_output, tuple):
        assert isinstance(test_output, tuple)
        assert len(test_output) == len(ref_output)
        for i in range(len(ref_output)):
            assert isinstance(test_output[i], type(ref_output[i]))
    else:
        assert isinstance(test_output, type(ref_output))


def test_atom_mask():
    """
    Test atom mask usage by relaxing only part of the model and
    expect that no unmasked hydrogen positions changed.
    """
    MASKED_RES_IDS = np.arange(1, 11)

    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    ref_coord = atoms.coord.copy()

    mask = np.isin(atoms.res_id, MASKED_RES_IDS)
    test_coord = hydride.relax_hydrogen(atoms, mask=mask)

    assert (test_coord[~mask] == ref_coord[~mask]).all()
    assert not (test_coord[mask] == ref_coord[mask]).all()


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "hydride"
__author__ = "Patrick Kunzmann, Jacob Marcel Anter"
__all__ = ["add_hydrogen"]

import numpy as np
import biotite.structure as struc
from .fragments import FragmentLibrary
from .names import AtomNameLibrary


def add_hydrogen(atoms, mask=None, fragment_library=None, name_library=None):
   """
   Add hydrogen atoms to a structure.

   The hydrogen atoms for each residue are placed directly behind the
   atoms from this residue.
   The function also tries to assign the correct atom name to each
   added hydrogen atom.

   Parameters
   ----------
   atoms : AtomArray, shape=(n,)
      The structure, where hydrogen atoms should be added.
      The structure must have an associated :class:`BondList`.
      The structure must also include the *charge* annotation array,
      depicting the formal charge for each atom.
      The structure must not have any annotated hydrogen atoms, yet.
   mask : ndarray, shape=(n,), dtype=bool, optional
      A boolean mask that is true for each atom, where hydrogen atoms
      should be added.
      By default, hydrogen atoms are added to all applicable atoms.
   fragment_library : FragmentLibrary
      The fragment library to use for hydrogen position estimation.
      By default :meth:`FragmentLibrary.standard_library()` is used,
      containing fragments from all molecules in the
      *RCSB* *Chemical Component Dictionary*.
   name_library : AtomNameLibrary
      The atom name library to use for hydrogen naming.
      By default :meth:`AtomNameLibrary.standard_library()` is used,
      containing atom names for the most prominent molecules including
      amino acids and nucleotides.
      For all other molecules the hydrogen atom names are guessed.

   Returns
   -------
   hydrogenated_atoms : AtomArray, shape=(p,)
      The atoms from the input `atoms` with additional hydrogen atoms.
      Although, the hydrogen positions are meaningful with respect to
      bond lengths and angles, the dihedral angles are not optimized.
   original_atom_mask : ndarray, shape=(p,)
      A boolean mask that is true for each atom in `hydrogenated_atoms`
      that originates from the input `atoms`.
      That means, that if the mask is applied to `hydrogenated_atoms`,
      the input structure is restored.
   """
   if mask is None:
      mask = np.ones(atoms.array_length(), dtype=bool)

   if fragment_library is None:
      fragment_library = FragmentLibrary.standard_library()
   if name_library is None:
      name_library = AtomNameLibrary.standard_library()

   if (atoms.element[mask] == "H").any():
      raise struc.BadStructureError(
         "Input structure already contains hydrogen atoms"
      )
   
   hydrogen_coord = fragment_library.calculate_hydrogen_coord(atoms, mask)

   # Count number of hydrogen atoms to be added
   count = 0
   p = 0
   for coord in hydrogen_coord:
      count += len(coord)
   
   # Create new empty AtomArray with an appropriate length for all
   # heavy and hydrogen atoms
   hydrogenated_atoms = struc.AtomArray(atoms.array_length() + count)
   original_atom_mask = np.zeros(hydrogenated_atoms.array_length(), dtype=bool)
   # Add all annotation categories of the original AtomArray
   for category in atoms.get_annotation_categories():
      if category not in hydrogenated_atoms.get_annotation_categories():
         hydrogenated_atoms.add_annotation(
            category, dtype=atoms.get_annotation(category).dtype
         )
   if atoms.box is not None:
      hydrogenated_atoms.box = atoms.box.copy()

   # Fill the combined AtomArray residue for residue
   # Stores covalent bonds between a heavy atom and its hydrogen atoms
   hydrogen_bonds = []
   residue_starts = struc.get_residue_starts(atoms, add_exclusive_stop=True)
   # Maps atom indices of the input AtomArray 
   # to indices of the hydrogenated AtomArray
   index_mapping = np.zeros(atoms.array_length(), dtype=np.uint32)
   p = 0
   for i in range(len(residue_starts) - 1):
      # Set annotation and coordinates from input AtomArray
      start = residue_starts[i]
      stop = residue_starts[i+1]
      res_length = stop - start
      index_mapping[start : stop] = np.arange(p, p + res_length)
      original_atom_mask[p : p + res_length] = True
      hydrogenated_atoms.coord[p : p + res_length] = atoms.coord[start : stop]
      for category in atoms.get_annotation_categories():
         hydrogenated_atoms.get_annotation(category)[p : p + res_length] \
                    = atoms.get_annotation(category)[start : stop]
      p += res_length
      # Set annotation and coordinates for hydrogen atoms
      for j in range(start, stop):
         hydrogen_coord_for_atom = hydrogen_coord[j]
         hydrogen_name_generator = name_library.generate_hydrogen_names(
            atoms.res_name[j], atoms.atom_name[j]
         )
         for coord in hydrogen_coord_for_atom:
            hydrogenated_atoms.coord[p] = coord
            hydrogenated_atoms.chain_id[p] = atoms.chain_id[j]
            hydrogenated_atoms.res_id[p] = atoms.res_id[j]
            hydrogenated_atoms.ins_code[p] = atoms.ins_code[j]
            hydrogenated_atoms.res_name[p] = atoms.res_name[j]
            hydrogenated_atoms.hetero[p] = atoms.hetero[j]
            hydrogenated_atoms.atom_name[p] = next(hydrogen_name_generator)
            hydrogenated_atoms.element[p] = "H"
            heavy_index = index_mapping[j]
            hydrogen_index = p
            hydrogen_bonds.append((heavy_index, hydrogen_index))
            p += 1
   
   # Add bonds to combined AtomArray
   original_bonds = atoms.bonds.as_array()
   bond_indices = index_mapping[original_bonds[:,:2]]
   heavy_bonds = np.stack(
      [
         bond_indices[:, 0],
         bond_indices[:, 1],
         # The bond types
         original_bonds[:,2]
      ],
      axis=-1
   )
   hydrogen_bonds = np.array(hydrogen_bonds, dtype=np.uint32).reshape(-1, 2)
   # All bonds to hydrogen atoms are single bonds
   hydrogen_bonds = np.stack(
      [
         hydrogen_bonds[:, 0],
         hydrogen_bonds[:, 1],
         np.ones(len(hydrogen_bonds), dtype=np.uint32)
      ],
      axis=-1
   )
   hydrogenated_atoms.bonds = struc.BondList(
      hydrogenated_atoms.array_length(),
      np.concatenate([heavy_bonds, hydrogen_bonds])
   )

   return hydrogenated_atoms, original_atom_mask


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
from os.path import join
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
import hydride
from .util import data_dir


def test_estimate_amino_acid_charges():
    """
    Test :func:`estimate_amino_acid_charges()` based on known charges
    at neutral pH.
    """

    ref_charges = {
        ("ARG", "NH2") :  1,
        ("ASP", "OD2") : -1,
        ("CYS",  "SG") :  0,
        ("GLU", "OE2") : -1,
        ("HIS", "ND1") :  0,
        ("LYS",  "NZ") :  1,
        ("TYR",  "OH") :  0,
    }
    
    mmtf_file = mmtf.MMTFFile.read(join(data_dir(), "1l2y.mmtf"))
    atoms = mmtf.get_structure(
        mmtf_file, model=1, include_bonds=True, extra_fields=["charge"]
    )
    heavy_atoms = atoms[atoms.element != "H"]
    
    test_charges = hydride.estimate_amino_acid_charges(heavy_atoms, 7.0)

    assert test_charges[0] == 1
    assert test_charges[-1] == -1
    for res_name, atom_name, test_charge in zip(
        heavy_atoms.res_name[1:-1],
        heavy_atoms.atom_name[1:-1],
        test_charges[1:-1]
    ):
        ref_charge = ref_charges.get((res_name, atom_name), 0)
        assert test_charge == ref_charge


# This source code is part of the Hydride package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "hydride"
__author__ = "Patrick Kunzmann"
__all__ = ["FragmentLibrary"]

from os.path import join, dirname, abspath
import warnings
import pickle
from biotite.structure.error import BadStructureError
from biotite.structure import BondType
import numpy as np


class FragmentLibrary:
    """
    A molecule fragment library for estimation of hydrogen positions.

    For each molecule added to the :class:`FragmentLibrary`,
    the molecule is split into fragments.
    Each fragment consists of

        - A central heavy atom,
        - bond order and position of its bonded heavy atoms and
        - and positions of bonded hydrogen atoms.
    
    The properties of the fragment (central atom element,
    central atom charge, order of connected bonds) are stored in
    a dictionary mapping these properties to heavy and hydrogen atom
    positions.

    If hydrogen atoms should be added to a target structure,
    the target structure is also split into fragments.
    Now the corresponding reference fragment in the library dictionary
    is accessed for each fragment.
    The corresponding atom coordinates of the reference fragment
    are superimposed [1]_ [2]_ onto the target fragment to obtain the
    hydrogen coordinates for the heavy atom.

    The constructor of this class creates an empty library.

    References
    ----------
    
    .. [1] W Kabsch,
       "A solution for the best rotation to relate two sets of vectors."
       Acta Cryst, 32, 922-923 (1976).
       
    .. [2] W Kabsch,
       "A discussion of the solution for the best rotation to relate
       two sets of vectors."
       Acta Cryst, 34, 827-828 (1978).
    """

    _std_library = None

    def __init__(self):
        self._frag_dict = {}
    

    @staticmethod
    def standard_library():
        """
        Get the standard :class:`FragmentLibrary`.
        The library contains fragments from all molecules in the
        *RCSB* *Chemical Component Dictionary*.

        Returns
        -------
        library : FragmentLibrary
            The standard library.
        """
        if FragmentLibrary._std_library is None:
            FragmentLibrary._std_library = FragmentLibrary()
            file_name = join(dirname(abspath(__file__)), "fragments.pickle")
            with open(file_name, "rb") as fragments_file:
                FragmentLibrary._std_library._frag_dict \
                    = pickle.load(fragments_file)
        return FragmentLibrary._std_library
    

    def add_molecule(self, molecule):
        """
        Add the fragments of a molecule to the library.

        Parameters
        ----------
        molecule : AtomArray
            A molecule, whose fragments should be added to the library.
            The structure must contain hydrogen atoms for each
            applicable heavy atom.
            The molecule must have an associated :class:`BondList`.
            The molecule must also include the *charge* annotation
            array, depicting the formal charge for each atom.
        """
        fragments = _fragment(molecule)
        for i, fragment in enumerate(fragments):
            if fragment is None:
                continue
            (
                central_element, central_charge, stereo, bond_types,
                center_coord, heavy_coord, hydrogen_coord
            ) = fragment
            # Translate the coordinates,
            # so the central heavy atom is at origin
            centered_heavy_coord = heavy_coord - center_coord
            centered_hydrogen_coord = hydrogen_coord - center_coord
            self._frag_dict[
                (central_element, central_charge, stereo, tuple(bond_types))
            ] = (
                # Information about the origin of the fragment
                # # for debugging purposes 
                molecule.res_name[i], molecule.atom_name[i],
                # The interesting information
                centered_heavy_coord, centered_hydrogen_coord
            )
            if stereo != 0:
                # Also include the opposite enantiomer in the library
                # by reflecting the coordinates along an arbitrary axis
                stereo *= -1
                refl_centered_heavy_coord = centered_heavy_coord.copy()
                refl_centered_hydrogen_coord = centered_hydrogen_coord.copy()
                refl_centered_heavy_coord[..., 0] *= -1
                refl_centered_hydrogen_coord[..., 0] *= -1
                self._frag_dict[(
                    central_element, central_charge, stereo, tuple(bond_types)
                )] = (
                    molecule.res_name[i], molecule.atom_name[i],
                    refl_centered_heavy_coord, refl_centered_hydrogen_coord
                )


    def calculate_hydrogen_coord(self, atoms, mask=None):
        """
        Estimate the hydrogen coordinates for each atom in a given
        structure/molecule.

        Parameters
        ----------
        atoms : AtomArray, shape=(n,)
            The structure to get the hydrogen atom positions for.
            The structure must not contain any hydrogen atoms.
            The structure must have an associated :class:`BondList`.
            The structure must also include the *charge* annotation
            array, depicting the formal charge for each atom.
        mask : ndarray, shape=(n,), dtype=bool
            A boolean mask that is true for each heavy atom, where
            corresponsing hydrogen atom positions should be calculated.
            By default, hydrogen atoms are calculated for all applicable
            atoms.

        Returns
        -------
        hydrogen_coord : list of (ndarray, shape=(k,3), dtype=np.float32), length=n
            A list of hydrogen coordinates for each atom in the input
            `atoms`.
            *k* is the number of hydrogen atoms for this atom.
            Each atom, that is not included in the input `mask` or which
            has no fitting fragment in the library,
            has no corresponding hydrogen coordinates, i.e. *k* is 0.
        """
        if mask is None:
            mask = np.ones(atoms.array_length(), dtype=bool)

        # The target and reference heavy atom coordinates
        # for each fragment
        sub_frag_center_coord = np.zeros(
            (atoms.array_length(), 3), dtype=np.float32
        )
        sub_frag_heavy_coord = np.zeros(
            (atoms.array_length(), 3, 3), dtype=np.float32
        )
        ref_frag_heavy_coord = np.zeros(
            (atoms.array_length(), 3, 3), dtype=np.float32
        )
        # The amount of hydrogens varies for each fragment
        # -> padding with NaN
        ref_frag_hydrogen_coord = np.full(
            (atoms.array_length(), 4, 3), np.nan, dtype=np.float32
        )

        # Fill the coordinate arrays
        fragments = _fragment(atoms, mask)
        for i, fragment in enumerate(fragments):
            if fragment is None:
                # This atom is not in mask
                continue
            (
                central_element, central_charge, stereo, bond_types,
                center_coord, heavy_coord, _
            ) = fragment
            sub_frag_center_coord[i] = center_coord
            sub_frag_heavy_coord[i] = heavy_coord
            # The hydrogen_coord can be ignored:
            # In the target structure are no hydrogen atoms
            hit = self._frag_dict.get(
                (central_element, central_charge, stereo, tuple(bond_types))
            )
            if hit is None:
                warnings.warn(
                    f"Missing fragment for atom '{atoms.atom_name[i]}' "
                    f"at position {i}"
                )
            else:
                _, _, ref_heavy_coord, ref_hydrogen_coord = hit
                ref_frag_heavy_coord[i] = ref_heavy_coord
                ref_frag_hydrogen_coord[i, :len(ref_hydrogen_coord)] \
                    = ref_hydrogen_coord

        # Translate the target coordinates,
        # so the central heavy atom is at origin
        # This has already been done for the reference atoms
        # in the 'add_molecule()' method
        sub_frag_heavy_coord -= sub_frag_center_coord[:, np.newaxis, :]
        # Get the rotation matrix required for superimposition of
        # the reference coord to the target coord 
        matrices = _get_rotation_matrices(
            sub_frag_heavy_coord, ref_frag_heavy_coord
        )
        # Rotate the reference hydrogen atoms, so they fit the
        # target heavy atoms
        sub_frag_hydrogen_coord = _rotate(ref_frag_hydrogen_coord, matrices)
        # Translate hydrogen atoms to the position of the
        # non-centered central heavy target atom
        sub_frag_hydrogen_coord += sub_frag_center_coord[:, np.newaxis, :]
        
        # Turn into list and remove NaN paddings
        sub_frag_hydrogen_coord = [
            # If the x-coordinate is NaN it is expected that
            # y and z are also NaN
            coord[~np.isnan(coord[:, 0])] for coord in sub_frag_hydrogen_coord
        ]

        return sub_frag_hydrogen_coord


def _fragment(atoms, mask=None):
    """
    Create fragments for the input structure/molecule.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,)
        The structure to be fragmented.
        The structure must have an associated :class:`BondList`.
        The structure must also include the *charge* annotation
        array, depicting the formal charge for each atom.
    mask : ndarray, shape=(n,), dtype=bool
        A boolean mask that is true for each heavy atom for which a
        fragment should be created.
    
    Returns
    -------
    fragments : list of tuple(str, int, int, ndarray, ndarray, ndarray), length=n
        The fragments.
        The tuple elements are

            #. the central atom element,
            #. the central atom charge,
            #. the enantiomer for stereocenter
               (``-1`` and ``1`` based on a custom nomenclature),
            #. the :class:`BondType` for each bonded heavy atom,
            #. the coordinates of the central atom,
            #. 3 coordinates of bonded heavy atoms (includes padding
               values, if there are not enough heavy atoms),
            #. the coordinates of bonded hydrogen atoms.
        
        ``None`` for each atom not included by the `mask`.
    """
    if mask is None:
        mask = np.ones(atoms.array_length(), dtype=bool)

    if atoms.bonds is None:
        raise BadStructureError(
            "The input structure must have an associated BondList"
        )

    fragments = [None] * atoms.array_length()
    
    all_bond_indices, all_bond_types = atoms.bonds.get_all_bonds()
    elements = atoms.element
    charges = atoms.charge
    coord = atoms.coord

    for i in range(atoms.array_length()):
        if not mask[i]:
            continue
        
        if elements[i] == "H":
            # Only create fragments for heavy atoms
            continue
        bond_indices = all_bond_indices[i]
        bond_types = all_bond_types[i]
        bond_indices = bond_indices[bond_indices != -1]
        bond_types = bond_types[bond_types != -1]

        heavy_mask = (elements[bond_indices] != "H")
        heavy_indices = bond_indices[heavy_mask]
        heavy_types = bond_types[heavy_mask]
        if (heavy_types == BondType.ANY).any():
            warnings.warn(
                f"Atom '{atoms.atom_name[i]}' in '{atoms.res_name[i]}' has an "
                f"undefined bond type and is ignored"
            )
            continue

        # Order the bonded atoms by their bond types
        # to remove atom order dependency in the matching step 
        order = np.argsort(heavy_types)
        heavy_indices = heavy_indices[order]
        heavy_types = heavy_types[order]

        hydrogen_mask = ~heavy_mask
        hydrogen_coord = coord[bond_indices[hydrogen_mask]]

        # Special handling of nitrogen as central atom:
        # There are cases where the free electron pair can form
        # a partial double bond.
        # Although the bond order is formally 1 in this case,
        # it would enforce planar hydrogen positionioning
        # Therefore, a partial double bond is handled as bond type 7
        if elements[i] == "N":
            for j, remote_index in enumerate(heavy_indices):
                if heavy_types[j] != 1:
                    # This handling only applies to single bonds
                    continue
                rem_bond_indices = all_bond_indices[remote_index]
                rem_bond_indices = rem_bond_indices[rem_bond_indices != -1]
                rem_bond_types = all_bond_types[remote_index]
                rem_bond_types = rem_bond_types[rem_bond_types != -1]
                for rem_rem_index, bond_type in zip(
                    rem_bond_indices, rem_bond_types
                ):
                    # If the adjacent atom has a double bond
                    # the partial double bond condition is fulfilled
                    if bond_type == BondType.AROMATIC_DOUBLE or \
                       bond_type == BondType.DOUBLE:
                            heavy_types[j] = 7

        n_heavy_bonds = np.count_nonzero(heavy_mask)
        if n_heavy_bonds == 0:
            # The orientation is arbitrary
            # -> The fragment coord is the coord of the central atom
            # 4 times repeated
            heavy_coord = np.repeat(coord[np.newaxis, i, :], 3, axis=0)
            stereo = 0
        elif n_heavy_bonds == 1:
            # Include one atom further away
            # to get an unambiguous fragment
            remote_index = heavy_indices[0]
            rem_bond_indices = all_bond_indices[remote_index]
            rem_bond_indices = rem_bond_indices[rem_bond_indices != -1]
            rem_heavy_mask = (elements[rem_bond_indices] != "H")
            rem_heavy_indices = rem_bond_indices[rem_heavy_mask]
            # Use the coord of any heavy atom bonded to the remote
            # atom
            rem_rem_index = rem_heavy_indices[0]
            # Include the directly bonded atom two times, to give it a
            # greater weight in superimposition
            heavy_coord = coord[[remote_index, remote_index, rem_rem_index]]
            stereo = 0
        elif n_heavy_bonds == 2:
            heavy_coord = coord[[heavy_indices[0], heavy_indices[1], i]]
            stereo = 0
        elif n_heavy_bonds == 3:
            heavy_coord = coord[heavy_indices]
            center = coord[i]
            # Determine the enantiomer of this stereocenter
            # For performance reasons, the result does not follow the
            # R/S nomenclature, but a custom -1/1 based one, which also
            # unambiguously identifies the enantiomer
            n = np.cross(heavy_coord[0] - center, heavy_coord[1] - center)
            stereo = int(np.sign(np.dot(heavy_coord[2] - center, n)))
        elif n_heavy_bonds == 4:
            # The fragment is irrelevant, as there is no bonded hydrogen
            # -> The fragment coord is the coord of the central atom
            # 4 times repeated
            heavy_coord = np.repeat(coord[np.newaxis, i, :], 3, axis=0)
            stereo = 0
        else:
            warnings.warn(
                f"Atom '{atoms.atom_name[i]}' in "
                f"'{atoms.res_name[i]}' has more than 4 bonds to "
                f"heavy atoms ({n_heavy_bonds}) and is ignored"
            )
            heavy_coord = np.repeat(coord[np.newaxis, i, :], 3, axis=0)
            hydrogen_coord = np.zeros((0, 3), dtype=np.float32)
            stereo = 0
        central_coord = coord[i]
        fragments[i] = (
            elements[i], charges[i], stereo, heavy_types,
            central_coord, heavy_coord, hydrogen_coord
        )
    return fragments


def _get_rotation_matrices(fixed, mobile):
    """
    Get the rotation matrices to superimpose the given mobile
    coordinates into the given fixed coordinates, minimizing the RMSD.

    Uses the *Kabsch* algorithm.

    Parameters
    ----------
    fixed : ndarray, shape=(m,n,3), dtype=np.float32
        The fixed coordinates.
    mobile : ndarray, shape=(m,n,3), dtype=np.float32
        The mobile coordinates.
    
    Returns
    -------
    matrices : ndarray, shape=(m,3,3), dtype=np.float32
        The rotation matrices.
    """
    # Calculate cross-covariance matrices
    cov = np.sum(fixed[:,:,:,np.newaxis] * mobile[:,:,np.newaxis,:], axis=1)
    v, s, w = np.linalg.svd(cov)
    # Remove possibility of reflected atom coordinates
    reflected_mask = (np.linalg.det(v) * np.linalg.det(w) < 0)
    v[reflected_mask, :, -1] *= -1
    matrices = np.matmul(v, w)
    return matrices


def _rotate(coord, matrices):
    """
    Apply a rotation on given coordinates.

    Parameters
    ----------
    coord : ndarray, shape=(m,n,3), dtype=np.float32
        The coordinates.
    matrices : ndarray, shape=(m,3,3), dtype=np.float32
        The rotation matrices.
    """
    return np.transpose(
        np.matmul(matrices, np.transpose(coord, axes=(0, 2, 1))),
        axes=(0, 2, 1)
    )